
/*

     Do a series of random walks on a 2d lattice.  Usage

	 ranwalk seed [cutoff]

     where `cutoff' is an optional (recommended!) cutoff applied
     after a given walk has done too many steps.  Otherwise the
     walk terminates when it returns to the origin, and the
     number of steps taken is printed.  It the walk was aborted
     because it took more than `cutoff' steps, -1 is output for
     that walk.  This program loops indefinitely.

*/


#include <stdio.h>
#include <math.h>

#define R_BIG 100

static char *usage        = "USAGE: ranwalk seed [cutoff]\n" ;

extern int  errno, sys_nerr ;
extern char *sys_errlist [] ;

void get_params (seed, do_cutoff, cutoff, argc, argv)

int argc ;
char *argv [] ;
int *seed, *do_cutoff, *cutoff ;

{
  if (argc < 2 || argc > 3)
  {
    printf (usage) ;
    exit (1) ;
  }
  *seed      = atoi (argv [1]) ;
  if (argc == 2)
  {
    *do_cutoff = 0 ;
  }
  else
  {
    *do_cutoff = 1 ;
    *cutoff    = atoi (argv [2]) ;
  }
}

static int YY [56], JJ, KK ;

void cra_init (seed)
int seed ;
{
  /* Initialise "Charlie's" random number generator. */

  int i ;

  JJ = 24 ;
  KK = 55 ;
  srandom (seed) ;
  for (i = 1; i < 55; i++)
    YY [i] = 0x3fffffff & random () ;
}

int cra_rand ()
{
  /*
    "Charlie's" random number generator.
    Restricted to range 0..2^30 - 1, to avoid overflows.
  */

  int result ;

  result  = 0x3fffffff & (YY [KK] + YY [JJ]) ;
  YY [KK] = result ;

  JJ-- ;
  KK-- ;
  if (JJ == 0)
    JJ = 55 ;
  if (KK == 0)
    KK = 55 ;

  return result ;
}

void main (argc, argv)
int argc ;
char *argv [] ;
{
  int seed, do_cutoff, cutoff ;

  setbuf (stdout, NULL) ;

  get_params (&seed, &do_cutoff, &cutoff, argc, argv) ;

  cra_init (seed) ;

  while (1)
  {
    int step = 0 ;
    int x = 0, y = 0 ;
    int returned = 0 ;
    int num_hits = 0 ;

    while (!returned && (!do_cutoff || step < cutoff))
    {
      float x_f   = (float) x ;
      float y_f   = (float) y ;
      float r_sqd = (x_f * x_f) + (y_f * y_f) ;

      if (r_sqd < (float) (R_BIG * R_BIG))
      {
	/* Single step */

	int u, b1, b2 ;

	/* This leads to bizzare non-random behaviour.
	u  = rand () ;
	b1 = (u & (1 << 14)) ;
	b2 = (u & (1 << 13)) ;
	*/
	/* Less bizzare, but obviously incorrect.
	u  = rand () ;
	b1 = (u & (1 << 14)) ;
	u  = rand () ;
	b2 = (u & (1 << 14)) ;
	*/
	u  = cra_rand () ;
	b1 = (u & (1 << 29)) ;
	u  = cra_rand () ;
	b2 = (u & (1 << 29)) ;
	/*
	u  = random () ;
	b1 = (u & (1 << 30)) ;
	u  = random () ;
	b2 = (u & (1 << 30)) ;
	*/
	if (b1 == 0)
	{
	  if (b2 == 0)
	    x++ ;
	  else
	    x-- ;
	}
	else
	{
	  if (b2 == 0)
	    y++ ;
	  else
	    y-- ;
	}
	step++ ;
      }
      else
      {
	/* Do a large block of steps, using asymptotic distribution */

	int t_step = (int) (r_sqd / R_BIG) ;
	int del_x, del_y ;

	do
	{
	  float del_sqd = - (t_step * 
	    log (((float) cra_rand ()) / (1<<30))) ;

	  float phi     = 2 * 3.1415927 * (((float) cra_rand ()) / (1<<30)) ;

	  float del = sqrt (del_sqd) ;

          /*
	  printf ("del_sqd = %f\n", del_sqd) ;
	  */
	  {
	    float d_x = del * cos (phi) ;

	    if (d_x >= 0)
	      /*
	      del_x = (int) (d_x + 0.5) ;
	      */
	      del_x = (int) (d_x) ;
	    else
	      /*
	      del_x = - (int) (- d_x + 0.5) ;
	      */
	      del_x = - (int) (- d_x) ;
	  }
	  {
	    float d_y = del * sin (phi) ;

	    if (d_y >= 0)
	      /*
	      del_y = (int) (d_y + 0.5) ;
	      */
	      del_y = (int) (d_y) ;
	    else
	      /*
	      del_y = - (int) (- d_y + 0.5) ;
	      */
	      del_y = - (int) (- d_y) ;
	  }
	  /*
	  printf ("t_step = %3d, del_x = %3d, del_y = %3d\n",
		   t_step, del_x, del_y) ;
	  */
	}
	while (((t_step + del_x + del_y) % 2) != 0) ;

        /*
	printf ("step = %3d, x = %3d, y = %3d\n", step, x, y) ;
	*/
	x += del_x ;
	y += del_y ;
	step += t_step ;
      }
      /*
      printf ("step = %3d, x = %3d, y = %3d\n", step, x, y) ;
      */
      returned = (x == 0) && (y == 0) ;
      if (returned)
	printf ("%d\n", step) ;
      else
      {
	if (do_cutoff && (step >= cutoff))
	{
	  printf ("abort x = %d, y = %d\n", x, y) ;
	}
      }
      num_hits++ ;
      /*
      if (num_hits > 1000000)
	printf ("step = %3d, x = %3d, y = %3d\n", step, x, y) ;
      */
    }
  }
}
